from typing import Literal, Optional

import pandas as pd

from bionty._public_ontology import PublicOntology
from bionty.dev._io import s3_bionty_assets
from bionty.entities._shared_docstrings import _doc_params, organism_removed


@_doc_params(doc_entities=organism_removed)
class Organism(PublicOntology):
    """Organism.

    1. Organism ontology
    Edits of terms are coordinated and reviewed on:
    https://www.ensembl.org/index.html

    Args:
        {doc_entities}
    """

    def __init__(
        self,
        organism: Optional[
            Literal["vertebrates", "bacteria", "fungi", "metazoa", "plants", "all"]
        ] = None,
        source: Optional[Literal["ensembl", "ncbitaxon"]] = None,
        version: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(organism=organism, source=source, version=version, **kwargs)

    def _load_df(self) -> pd.DataFrame:
        if self.source == "ensembl":
            if not self._local_parquet_path.exists():
                # try to download from s3
                s3_bionty_assets(
                    filename=self._parquet_filename,
                    assets_base_url="s3://bionty-assets",
                    localpath=self._local_parquet_path,
                )

            # try to download from original url
            if not self._local_parquet_path.exists():
                self._url_download(self._url, self._local_ontology_path)  # type:ignore
                df = pd.read_csv(
                    self._local_ontology_path,
                    sep="\t",
                    index_col=False,  # type:ignore
                )
                df.rename(
                    columns={
                        "#name": "name",
                        "species": "scientific_name",
                        "taxonomy_id": "ontology_id",
                    },
                    inplace=True,
                )
                df["name"] = df["name"].str.lower()
                df["ontology_id"] = "NCBITaxon:" + df["ontology_id"].astype(str)
                df.to_parquet(self._local_parquet_path)
                return df
            else:
                return pd.read_parquet(self._local_parquet_path)
        else:
            return super()._load_df()

    def df(self) -> pd.DataFrame:
        """Pandas DataFrame of the ontology.

        Returns:
            A Pandas DataFrame of the ontology.

        Examples:
            >>> import bionty as bt
            >>> bt.Organism().df()
        """
        return self._df.set_index("name")



from typing import Dict, Iterable, Literal, Optional

import pandas as pd
from lamin_utils import logger

from bionty._public_ontology import PublicOntology
from bionty._settings import settings
from bionty.dev._io import s3_bionty_assets

from ._organism import Organism
from ._shared_docstrings import _doc_params, doc_entites


@_doc_params(doc_entities=doc_entites)
class Gene(PublicOntology):
    """Gene.

    1. Ensembl
    Edits of terms are coordinated and reviewed on:
    https://www.ensembl.org/

    Args:
        {doc_entities}

    Notes:
        Biotypes: https://www.ensembl.org/info/genome/genebuild/biotypes.html
        Gene Naming: https://www.ensembl.org/info/genome/genebuild/gene_names.html
    """

    def __init__(
        self,
        organism: Optional[
            Literal["human", "mouse", "saccharomyces cerevisiae"]
        ] = None,
        source: Optional[Literal["ensembl"]] = None,
        version: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(
            source=source,
            version=version,
            organism=organism,
            **kwargs,
        )

    def map_legacy_ids(self, values: Iterable):
        """Convert legacy ids to current ids."""
        if self.source != "ensembl":
            raise NotImplementedError
        if isinstance(values, str):
            values = [values]
        ensembl = EnsemblGene(organism=self.organism, version=self.version)
        legacy_df_filename = f"df-legacy_{self.organism}__{self.source}__{self.version}__{self.__class__.__name__}.parquet"
        legacy_df_localpath = settings.dynamicdir / legacy_df_filename
        s3_bionty_assets(
            filename=legacy_df_filename,
            assets_base_url="s3://bionty-assets",
            localpath=legacy_df_localpath,
        )
        try:
            results = pd.read_parquet(legacy_df_localpath)
        except FileNotFoundError:
            raise NotImplementedError from None
        results = results[results.old_stable_id.isin(values)].copy()
        return ensembl._process_convert_result(results, values=values)


class EnsemblGene:
    def __init__(self, organism: str, version: str) -> None:
        """Ensembl Gene mysql.

        Args:
            organism: a bionty.Organism object
            version: name of the ensembl DB version, e.g. "release-110"
        """
        self._import()
        import mysql.connector as sql
        from sqlalchemy import create_engine

        self._organism = (
            Organism(version=version).lookup().dict().get(organism)  # type:ignore
        )
        self._url = (
            f"mysql+mysqldb://anonymous:@ensembldb.ensembl.org/{self._organism.core_db}"
        )
        self._engine = create_engine(url=self._url)

    def _import(self):
        try:
            import mysql.connector as sql
            from sqlalchemy import create_engine
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "To query from the Ensembl database, please run `pip install"
                " sqlalchemy mysqlclient mysql-connector-python`"
            ) from None

    def external_dbs(self):
        return pd.read_sql("SELECT * FROM external_db", con=self._engine)

    def download_df(self, external_db_names: Optional[Dict] = None) -> pd.DataFrame:
        """Fetch gene table from Ensembl mysql database.

        Args:
            external_db_names: {external database name : df column name}, see `.external_dbs()`
                Default is {"EntrezGene": "ensembl_gene_id"}.
        """
        query_core = """
        SELECT gene.stable_id, xref.display_label, gene.biotype, gene.description, external_synonym.synonym
        FROM gene
        LEFT JOIN xref ON gene.display_xref_id = xref.xref_id
        LEFT JOIN external_synonym ON gene.display_xref_id = external_synonym.xref_id
        """

        entrez_name = {"EntrezGene": "ensembl_gene_id"}
        if external_db_names is None:
            external_db_names = entrez_name
        if entrez_name not in list(external_db_names.keys()):
            external_db_names.update(entrez_name)
        external_db_names_str = ", ".join(
            [f"'{name}'" for name in external_db_names.keys()]
        )

        query_external = f"""
        SELECT gene.stable_id, object_xref.xref_id, xref.dbprimary_acc, external_db.db_name
        FROM gene
        LEFT JOIN object_xref ON gene.gene_id = object_xref.ensembl_id
        LEFT JOIN xref ON object_xref.xref_id = xref.xref_id
        LEFT JOIN external_db ON xref.external_db_id = external_db.external_db_id
        WHERE object_xref.ensembl_object_type = 'Gene' AND external_db.db_name IN ({external_db_names_str}) # noqa
        """

        # Query for the basic gene annotations:
        results_core = pd.read_sql(query_core, con=self._engine)
        logger.info("fetching records from the core DB...")

        # aggregate metadata based on ensembl stable_id
        results_core_group = results_core.groupby("stable_id").agg(
            {
                "display_label": "first",
                "biotype": "first",
                "description": "first",
                "synonym": lambda x: "|".join([i for i in set(x) if i is not None]),
            }
        )

        # Query for external ids:
        results_external = pd.read_sql(query_external, con=self._engine)
        logger.info("fetching records from the external DBs...")

        def add_external_db_column(df: pd.DataFrame, ext_db: str, df_col: str):
            # ncbi_gene_id
            ext = (
                results_external[results_external["db_name"] == ext_db]
                .drop_duplicates(["stable_id", "dbprimary_acc"])
                .drop(columns=["xref_id", "db_name"])
            )
            ext.rename(columns={"dbprimary_acc": df_col}, inplace=True)
            ext = ext.set_index("stable_id")
            dup = ext[ext.index.duplicated(keep=False)]
            if dup.shape[0] > 0:
                logger.warning(
                    f"duplicated #rows ensembl_gene_id with {df_col}: {dup.shape[0]}"
                )
            df_merge = df.merge(ext, left_index=True, right_index=True, how="outer")
            return df_merge

        df = add_external_db_column(
            df=results_core_group, ext_db="EntrezGene", df_col="ncbi_gene_id"
        )

        for ext_db, df_col in external_db_names.items():
            if ext_db == "EntrezGene":
                continue
            df = add_external_db_column(df=df, ext_db=ext_db, df_col=df_col)

        df = df.reset_index()
        df.rename(
            columns={
                "stable_id": "ensembl_gene_id",
                "display_label": "symbol",
                "synonym": "synonyms",
            },
            inplace=True,
        )
        df_res = df.loc[
            :,
            [
                "ensembl_gene_id",
                "symbol",
                "ncbi_gene_id",
                "biotype",
                "description",
                "synonyms",
            ],
        ].copy()
        for col in df.columns:
            if col not in df_res.columns:
                df_res[col] = df[col]
        df_res = df_res[~df_res["ensembl_gene_id"].isnull()]
        df_res = df_res[~df_res["ensembl_gene_id"].isna()]

        # if stable_id is not ensembl_gene_id, keep a stable_id column
        if not any(df_res["ensembl_gene_id"].str.startswith("ENS")):
            logger.warning("no ensembl_gene_id found, writing to table_id column.")
            df_res.insert(0, "stable_id", df_res.pop("ensembl_gene_id"))
            df_res = df_res.sort_values("stable_id").reset_index(drop=True)
        else:
            df_res = df_res[df_res["ensembl_gene_id"].str.startswith("ENS")]
            df_res = df_res.sort_values("ensembl_gene_id").reset_index(drop=True)

        logger.success(f"downloaded Gene table containing {df_res.shape[0]} entries.")

        return df_res

    def download_legacy_ids_df(self, df: pd.DataFrame, col: Optional[str] = None):
        col = "ensembl_gene_id" if col is None else col
        current_ids = tuple(df[col])
        results = pd.read_sql(
            "SELECT * FROM stable_id_event JOIN mapping_session USING"
            " (mapping_session_id) WHERE type = 'gene' AND new_stable_id IN"
            f" {current_ids} AND score > 0 AND old_stable_id != new_stable_id",
            con=self._engine,
        )
        return results

    def _process_convert_result(
        self,
        results: pd.DataFrame,
        values: Iterable,
    ):
        # unique mappings
        mapper = (
            results.drop_duplicates(["old_stable_id"], keep=False)
            .set_index("old_stable_id")["new_stable_id"]
            .to_dict()
        )
        # ambiguous mappings
        ambiguous = (
            results[~results["old_stable_id"].isin(mapper)][
                ["old_stable_id", "new_stable_id"]
            ]
            .groupby("old_stable_id", group_keys=False)["new_stable_id"]
            .apply(list)
            .to_dict()
        )
        # unmappables
        unmapped = set(values).difference(results["old_stable_id"])
        return {"mapped": mapper, "ambiguous": ambiguous, "unmapped": list(unmapped)}

    def map_legacy_ids(self, values: Iterable, df: pd.DataFrame):
        if isinstance(values, str):
            legacy_genes = f"('{values}')"
            values = [values]
        else:
            legacy_genes = tuple(values)  # type:ignore
        if len(legacy_genes) == 1:
            legacy_genes = f"('{legacy_genes[0]}')"

        current_ids = tuple(df.ensembl_gene_id)
        # query the ensembl mysql db
        results = pd.read_sql(
            "SELECT * FROM stable_id_event JOIN mapping_session USING"
            " (mapping_session_id) WHERE type = 'gene' AND old_stable_id IN"
            f" {legacy_genes} AND new_stable_id IN {current_ids} AND old_stable_id !="
            " new_stable_id",
            con=self._engine,
        )
        return self._process_convert_result(results, values)



import os
from pathlib import Path
from typing import Union

import requests  # type:ignore
import yaml  # type:ignore
from rich.progress import Progress

from bionty._settings import settings


def load_yaml(
    filename: Union[str, Path], convert_dates: bool = True
):  # pragma: no cover
    with open(filename) as f:
        return yaml.safe_load(f)


def write_yaml(
    data: dict,
    filename: Union[str, Path],
    sort_keys: bool = False,
    default_flow_style: bool = False,
):  # pragma: no cover
    with open(filename, "w") as f:
        yaml.dump(
            data,
            f,
            sort_keys=sort_keys,
            default_flow_style=default_flow_style,
        )


def url_download(
    url: str, localpath: Union[str, Path, None] = None, block_size: int = 1024, **kwargs
) -> Union[str, Path, None]:
    """Downloads a file to a specified path.

    Args:
        url: The URL to download.
        localpath: The path to download the file to.
        block_size: Buffer size in bytes for sending a file-like message body.
        **kwargs: Keyword arguments are passed to 'requests'

    Returns:
        The localpath file is downloaded to

    Raises:
        HttpError: If the request response is not 200 and OK.
    """
    try:
        response = requests.get(url, stream=True, allow_redirects=True, **kwargs)
        response.raise_for_status()

        total_content_length = int(response.headers.get("content-length", 0))
        if localpath is None:
            localpath = url.split("/")[-1]

        if total_content_length > 5000000:
            with Progress(refresh_per_second=10, transient=True) as progress:
                task = progress.add_task(
                    "[red]downloading...", total=total_content_length
                )

                with open(localpath, "wb") as file:
                    for data in response.iter_content(block_size):
                        file.write(data)
                        progress.update(task, advance=block_size)
                # force the progress bar to 100% at the end
                progress.update(task, completed=total_content_length, refresh=True)
        else:
            with open(localpath, "wb") as file:
                for data in response.iter_content(block_size):
                    file.write(data)

        return localpath

    except requests.exceptions.HTTPError as err:
        raise err


def s3_bionty_assets(
    filename: str, localpath: Path = None, assets_base_url: str = "s3://bionty-assets"
):
    """Synchronizes a S3 file path with local file storage.

    If the file does not exist locally it gets downloaded to datasetdir/filename or the passed localpath.
    If the file does not exist on S3, the file does not get synchronized, no erroring.

    Args:
        filename: The suffix of the assets_base_url.
        localpath: Local base path of the file to sync.
        assets_base_url: The S3 base URL. Prefix of the filename.

    Returns:
        A Path object of the synchronized path.
    """
    import botocore.session as session
    from botocore.config import Config

    if localpath is None:
        localpath = settings.datasetdir / filename
    elif localpath.is_dir():
        localpath = localpath / filename

    bucket = assets_base_url.replace("s3://", "")
    s3_client = session.get_session().create_client(
        "s3", config=Config(signature_version=session.UNSIGNED)
    )

    try:
        s3_object = s3_client.get_object(Bucket=bucket, Key=filename)
    except s3_client.exceptions.NoSuchKey:
        return localpath

    cloud_mts = s3_object["LastModified"].timestamp()
    total_content_length = int(s3_object["ContentLength"])

    CHUNK_SIZE = 64 * 1024

    if not localpath.exists() or cloud_mts > localpath.stat().st_mtime:  # type: ignore
        localpath.parent.mkdir(parents=True, exist_ok=True)
        stream = s3_object["Body"]
        if total_content_length > 5000000:
            with Progress(refresh_per_second=10, transient=True) as progress:
                task = progress.add_task(
                    "[red]downloading...", total=total_content_length
                )
                try:
                    with localpath.open(mode="wb") as f:
                        for chunk in iter(lambda: stream.read(CHUNK_SIZE), b""):
                            f.write(chunk)
                            progress.update(task, advance=CHUNK_SIZE)
                except Exception as e:
                    if localpath.exists():
                        localpath.unlink()
                    raise e
                # force the progress bar to 100% at the end
                progress.update(task, completed=total_content_length, refresh=True)
        else:
            try:
                with localpath.open(mode="wb") as f:
                    for chunk in iter(lambda: stream.read(CHUNK_SIZE), b""):
                        f.write(chunk)
            except Exception as e:
                if localpath.exists():
                    localpath.unlink()
                raise e
        os.utime(localpath, times=(cloud_mts, cloud_mts))

    return localpath



from pathlib import Path

import pytest
from bionty.dev._io import url_download


@pytest.fixture
def local(tmp_path):
    url = "https://bionty-assets.s3.amazonaws.com/bfxpipelines.json"
    localpath = tmp_path / Path(url).name
    yield localpath, url
    if localpath.exists():
        localpath.unlink()


def test_url_download(local):
    localpath = local[0]
    url = local[1]
    assert not localpath.exists()

    downloaded_path = Path(url_download(url=url, localpath=localpath))
    assert downloaded_path.exists()



from typing import Literal, Optional

from bionty._public_ontology import PublicOntology

from ._shared_docstrings import _doc_params, doc_entites


@_doc_params(doc_entities=doc_entites)
class Protein(PublicOntology):
    """Protein.

    1. Uniprot
    Edits of terms are coordinated and reviewed on:
    https://www.uniprot.org/

    Args:
        {doc_entities}
    """

    def __init__(
        self,
        organism: Optional[Literal["human", "mouse"]] = None,
        source: Optional[Literal["uniprot"]] = None,
        version: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(source=source, version=version, organism=organism, **kwargs)



import bionty as bt
import pandas as pd


def test_cellmarker_cellmarker_inspect_name_human():
    df = pd.DataFrame(
        index=["CCR7", "CD69", "CD8", "CD45RA", "This protein does not exist"]
    )

    cm = bt.CellMarker(source="cellmarker")
    curated = cm.inspect(df.index, field=cm.name)

    assert curated["validated"] == ["CD69", "CD8", "CD45RA"]
    assert curated["non_validated"] == ["CCR7", "This protein does not exist"]


def test_cellmarker_cellmarker_inspect_name_mouse():
    df = pd.DataFrame(
        index=["Tcf4", "Cd36", "Cd34", "Lgr6", "This protein does not exist"]
    )

    cm = bt.CellMarker(source="cellmarker", organism="mouse")
    inspected_df = cm.inspect(df.index, field=cm.name, return_df=True)

    inspect = inspected_df["__validated__"].reset_index(drop=True)
    expected_series = pd.Series([True, False, True, True, False])

    assert inspect.equals(expected_series)



"""Bionty.

Import the package::

   import bionty as bt

Entities:

.. autosummary::
   :toctree: .

   Organism
   Gene
   Protein
   CellMarker
   CellType
   CellLine
   Tissue
   Disease
   Phenotype
   Pathway
   ExperimentalFactor
   DevelopmentalStage
   Drug
   Ethnicity
   BFXPipeline
   BioSample

Base model of entity classes:

.. autosummary::
   :toctree: .

   PublicOntology
   PublicOntologyField

PublicOntology sources:

.. autosummary::
   :toctree: .

    display_available_sources
    display_currently_used_sources
    reset_sources
    settings

External API:

.. autosummary::
   :toctree: .

   Ontology
"""

__version__ = "0.36.0"  # denote release candidate for 0.1.0 with 0.1rc1

# prints warning of python versions
from lamin_utils import py_version_warning

py_version_warning("3.8", "3.10")

from ._sync_sources import sync_sources

sync_sources()

# dynamic classes
from . import dev
from ._display_sources import display_available_sources, display_currently_used_sources
from ._ontology import Ontology

# tools
from ._public_ontology import PublicOntology, PublicOntologyField
from ._settings import settings

# sources
from .dev._handle_sources import reset_sources
from .entities._bfxpipeline import BFXPipeline
from .entities._biosample import BioSample
from .entities._cellline import CellLine
from .entities._cellmarker import CellMarker
from .entities._celltype import CellType
from .entities._developmentalstage import DevelopmentalStage
from .entities._disease import Disease
from .entities._drug import Drug
from .entities._ethnicity import Ethnicity
from .entities._experimentalfactor import ExperimentalFactor
from .entities._gene import Gene
from .entities._organism import Organism
from .entities._pathway import Pathway
from .entities._phenotype import Phenotype
from .entities._protein import Protein
from .entities._tissue import Tissue

# backward compat
Entity = PublicOntology
Bionty = PublicOntology
Readout = ExperimentalFactor
Species = Organism



import bionty as bt
import pytest
from bionty._settings import settings


def test_unavailable_sources():
    with pytest.raises(ValueError):
        bt.CellType(source="random")


def test_reset_sources(monkeypatch):
    import shutil

    monkeypatch.setattr("builtins.input", lambda _: "y")

    shutil.copyfile(
        settings.current_sources.as_posix(), settings.lamindb_sources.as_posix()
    )
    bt.reset_sources()

    settings.current_sources.unlink()
    settings.local_sources.unlink()
    bt.reset_sources()


def test_diff_successful():
    disease_bt_1 = bt.Disease(source="mondo", version="2023-04-04")
    disease_bt_2 = bt.Disease(source="mondo", version="2023-02-06")

    new_entries, modified_entries = disease_bt_1.diff(disease_bt_2)
    assert len(new_entries) == 819
    assert len(modified_entries) == 249


def test_diff_value_errors():
    # Two different PublicOntology object types
    disease_bt = bt.Disease()
    phenotype_bt = bt.Phenotype()
    with pytest.raises(ValueError):
        disease_bt.diff(phenotype_bt)

    # Different sources
    disease_bt_1 = bt.Disease(source="mondo")
    disease_bt_2 = bt.Disease(source="doid")
    with pytest.raises(ValueError):
        disease_bt_1.diff(disease_bt_2)

    # Same version
    disease_bt_3 = bt.Disease(source="mondo", version="2023-04-04")
    disease_bt_4 = bt.Disease(source="mondo", version="2023-04-04")
    with pytest.raises(ValueError):
        disease_bt_3.diff(disease_bt_4)



import bionty as bt
import pandas as pd


def test_hsapdv_developmentalstage_inspect_name():
    df = pd.DataFrame(
        index=[
            "blastula stage",
            "Carnegie stage 03",
            "neurula stage",
            "organogenesis stage",
            "This developmental stage does not exist",
        ]
    )

    ds = bt.DevelopmentalStage(source="hsapdv")
    inspected_df = ds.inspect(df.index, field=ds.name, return_df=True)

    inspect = inspected_df["__validated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)



import nox
from laminci import move_built_docs_to_docs_slash_project_slug
from laminci.nox import build_docs, run_pre_commit

nox.options.default_venv_backend = "none"


@nox.session
def lint(session: nox.Session) -> None:
    run_pre_commit(session)


@nox.session
def docs(session: nox.Session):
    build_docs(session, strict=True)


@nox.session
@nox.parametrize("group", ["bionty-unit", "bionty-docs"])
def build(session: nox.Session, group: str):
    session.run(*"pip install -e .[dev]".split())
    coverage_args = "--cov=bionty --cov-append --cov-report=term-missing"
    if group == "bionty-unit":
        session.run(*f"pytest {coverage_args} ./tests".split())
    elif group == "bionty-docs":
        session.run(*f"pytest -s {coverage_args} ./docs/guide".split())
        docs(session)
        move_built_docs_to_docs_slash_project_slug()



import os
import tempfile
from pathlib import Path

import pytest
from bionty._settings import settings
from bionty.dev._handle_sources import (
    add_records_to_existing_dict,
    parse_currently_used_sources,
    parse_sources_yaml,
    records_diff_btw_yamls,
    reset_sources,
    update_local_from_public_sources_yaml,
)
from bionty.dev._io import load_yaml, write_yaml


@pytest.fixture(scope="function")
def versions_yaml_replica():
    input_file_content = """
    version: "0.2.0"
    Organism:
      ensembl:
        all:
          release-108:
            url: https://ftp.ensembl.org/pub/release-108/species_EnsemblVertebrates.txt
        name: Ensembl
        website: https://www.ensembl.org/index.html
    Gene:
      ensembl:
        human:
          release-108:
            url: https://ftp.ensembl.org/pub/release-108/mysql/homo_sapiens_core_108_38/
          release-107:
            url: https://ftp.ensembl.org/pub/release-107/mysql/homo_sapiens_core_107_38/
        mouse:
          release-108:
            url: https://ftp.ensembl.org/pub/release-108/mysql/mus_musculus_core_108_39/
        name: Ensembl
        website: https://www.ensembl.org/index.html
    CellType:
      cl:
        name: Cell Ontology
        website: https://obophenotype.github.io/cell-ontology/
        all:
          2023-02-15:
            url: http://purl.obolibrary.org/obo/cl/releases/2023-02-15/cl-base.owl
            md5: 9331a6a029cb1863bd0584ab41508df7
          2022-08-16:
            url: http://purl.obolibrary.org/obo/cl/releases/2022-08-16/cl.owl
            md5: d0655766574e63f3fe5ed56d3c030880
    """
    with tempfile.NamedTemporaryFile(mode="w+", delete=False) as f:
        f.write(input_file_content)
        f.flush()
        yield f.name

    Path(f.name).unlink()


@pytest.fixture(scope="function")
def new_versions_yaml_replica():
    input_file_content = """
    version: "0.2.0"
    Organism:
      ensembl:
        all:
          release-108:
            url: https://ftp.ensembl.org/pub/release-108/species_EnsemblVertebrates.txt
        name: Ensembl
        website: https://www.ensembl.org/index.html
        new-organism:
          release-x:
            url: new-organism-source-link
    Gene:
      ensembl:
        human:
          release-108:
            url: https://ftp.ensembl.org/pub/release-108/mysql/homo_sapiens_core_108_38/
          release-107:
            url: https://ftp.ensembl.org/pub/release-107/mysql/homo_sapiens_core_107_38/
        mouse:
          release-108:
            url: https://ftp.ensembl.org/pub/release-108/mysql/mus_musculus_core_108_39/
        name: Ensembl
        website: https://www.ensembl.org/index.html
      new-source:
        human:
          release-x:
            url: new-gene-source-link
    CellType:
      cl:
        name: Cell Ontology
        website: https://obophenotype.github.io/cell-ontology/
        all:
          new-version:
            url: new-cell-type-source
            md5: new-md5
          2023-02-15:
            url: http://purl.obolibrary.org/obo/cl/releases/2023-02-15/cl-base.owl
            md5: 9331a6a029cb1863bd0584ab41508df7
          2022-08-16:
            url: http://purl.obolibrary.org/obo/cl/releases/2022-08-16/cl.owl
            md5: d0655766574e63f3fe5ed56d3c030880
    """
    with tempfile.NamedTemporaryFile(mode="w+", delete=False) as f:
        f.write(input_file_content)
        f.flush()
        yield f.name

    Path(f.name).unlink()


@pytest.fixture(scope="function")
def current_yaml_replica():
    input_file_content = """
    Organism:
      all:
        ensembl: release-108
    """
    with tempfile.NamedTemporaryFile(mode="w+", delete=False) as f:
        f.write(input_file_content)
        f.flush()
        yield f.name

    Path(f.name).unlink()


def test_parse_versions_yaml(versions_yaml_replica):
    parsed_df = parse_sources_yaml(versions_yaml_replica)
    assert parsed_df.shape == (6, 8)
    assert all(
        parsed_df["entity"].values
        == ["Organism", "Gene", "Gene", "Gene", "CellType", "CellType"]
    )
    assert all(
        parsed_df["organism"].values == ["all", "human", "human", "mouse", "all", "all"]
    )
    assert all(
        parsed_df["source"].values
        == ["ensembl", "ensembl", "ensembl", "ensembl", "cl", "cl"]
    )


def test_parse_current_versions(versions_yaml_replica):
    expected = {
        "Organism": {"all": {"ensembl": "release-108"}},
        "Gene": {
            "human": {"ensembl": "release-108"},
            "mouse": {"ensembl": "release-108"},
        },
        "CellType": {"all": {"cl": "2023-02-15"}},
    }

    assert parse_currently_used_sources(versions_yaml_replica) == expected


def test_add_records_to_existing_dict(new_versions_yaml_replica, versions_yaml_replica):
    expected = [
        {
            "entity": "Organism",
            "source": "ensembl",
            "organism": "new-organism",
            "version": "release-x",
            "url": "new-organism-source-link",
            "md5": "",
            "source_name": "Ensembl",
            "source_website": "https://www.ensembl.org/index.html",
        },
        {
            "entity": "Gene",
            "source": "new-source",
            "organism": "human",
            "version": "release-x",
            "url": "new-gene-source-link",
            "md5": "",
            "source_name": "",
            "source_website": "",
        },
        {
            "entity": "CellType",
            "source": "cl",
            "organism": "all",
            "version": "new-version",
            "url": "new-cell-type-source",
            "md5": "new-md5",
            "source_name": "Cell Ontology",
            "source_website": "https://obophenotype.github.io/cell-ontology/",
        },
    ]

    records = records_diff_btw_yamls(new_versions_yaml_replica, versions_yaml_replica)

    assert records == expected

    updated_dict = add_records_to_existing_dict(
        records, load_yaml(versions_yaml_replica)
    )
    assert updated_dict.get("Organism").get("ensembl").get("new-organism").get(
        "release-x"
    ) == {"url": "new-organism-source-link", "md5": ""}
    assert updated_dict.get("Gene").get("new-source").get("human").get("release-x") == {
        "url": "new-gene-source-link",
        "md5": "",
    }
    assert updated_dict.get("CellType").get("cl").get("all").get("new-version") == {
        "url": "new-cell-type-source",
        "md5": "new-md5",
    }


def test_update_local_from_public_sources_yaml():
    local_dict = load_yaml(settings.local_sources)
    local_dict.pop("Organism")
    write_yaml(local_dict, settings.local_sources)
    update_local_from_public_sources_yaml()


def test_reset_sources(monkeypatch):
    monkeypatch.setattr("builtins.input", lambda _: "y")
    import shutil

    shutil.copyfile(
        settings.current_sources.as_posix(), settings.lamindb_sources.as_posix()
    )
    reset_sources()



from typing import Literal, Optional

from bionty._public_ontology import PublicOntology

from ._shared_docstrings import _doc_params, doc_entites


@_doc_params(doc_entities=doc_entites)
class Phenotype(PublicOntology):
    """Phenotype.

    1. Human Phenotype Ontology
    Edits of terms are coordinated and reviewed on:
    https://hpo.jax.org/app/

    2. Mammalian Phenotype Ontology
    Edits of terms are coordinated and reviewed on:
    https://github.com/mgijax/mammalian-phenotype-ontology

    3. Zebrafish Phenotype Ontology
    Edits of terms are coordinated and reviewed on:
    https://github.com/obophenotype/zebrafish-phenotype-ontology

    4.Phecodes ICD10 map
    Website:
    https://phewascatalog.org/phecodes_icd10

    Args:
        {doc_entities}
    """

    def __init__(
        self,
        organism: Optional[Literal["human", "mammalian", "zebrafish", "all"]] = None,
        source: Optional[Literal["hp", "phe", "mp", "zp", "pato"]] = None,
        version: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            source=source,
            version=version,
            organism=organism,
            include_id_prefixes={
                "hp": ["HP"],
                "mp": ["MP"],  # mp might require an exclusion prefix for mpath
                "zp": ["ZP"],
                "pato": ["PATO"],
            },
            **kwargs,
        )



from __future__ import annotations

import logging
from typing import (
    TYPE_CHECKING,
    Iterable,
    Literal,
)

import numpy as np
import pandas as pd
from lamin_utils import logger
from lamin_utils._lookup import Lookup

from ._ontology import Ontology
from ._settings import check_datasetdir_exists, check_dynamicdir_exists, settings
from .dev._handle_sources import LAMINDB_INSTANCE_LOADED
from .dev._io import s3_bionty_assets, url_download
from .dev._md5 import verify_md5

if TYPE_CHECKING:
    from pathlib import Path

    from .dev import InspectResult


def encode_filenames(
    organism: str, source: str, version: str, entity: PublicOntology | str
) -> tuple[str, str]:
    """Encode names of the cached files."""
    if isinstance(entity, PublicOntology):
        entity_name = entity.__class__.__name__
    else:
        entity_name = entity
    parquet_filename = f"df_{organism}__{source}__{version}__{entity_name}.parquet"
    ontology_filename = (
        f"ontology_{organism}__{source}__{version}__{entity_name}".replace(" ", "_")
    )

    return parquet_filename, ontology_filename


class PublicOntology:
    """PublicOntology object."""

    def __init__(
        self,
        source: str | None = None,
        version: str | None = None,
        organism: str | None = None,
        *,
        include_id_prefixes: dict[str, list[str]] | None = None,
        **kwargs,
    ):
        # backward compat for species -> organism
        if organism is None and kwargs.get("species") is not None:
            organism = kwargs.get("species")
        self._fetch_sources()
        try:
            # match user input organism, source and version with currently used sources
            current = self._match_sources(
                self._current_sources,
                source=source,
                version=version,
                organism=organism,
            )
            source = current.get("source")
            version = current.get("version")
            organism = current.get("organism")
        except ValueError:
            if LAMINDB_INSTANCE_LOADED():
                logger.warning("loading non-default source inside a LaminDB instance")
                # fmt: off
                logger.hint(
                    f"please consider:\n"
                    f"    close your instance via `lamin close` and use Bionty stand alone\n"
                    f"    OR\n"
                    f"    modify currently_used {self.__class__.__name__} source in `lnschema_bionty.PublicSource`"
                )
                # fmt: on

        # search in all available sources to get url and md5
        self._source_record = self._match_sources(
            self._all_sources,
            source=source,
            version=version,
            organism=organism,
        )

        self._organism = self._source_record["organism"]
        self._source = self._source_record["source"]
        self._version = self._source_record["version"]

        self._set_file_paths()
        self.include_id_prefixes = include_id_prefixes

        # df is only read into memory at the init to improve performance
        df = self._load_df()
        # self._df has no index
        if df.index.name is not None:
            df = df.reset_index()
        self._df: pd.DataFrame = df

        # set column names/fields as attributes
        for col_name in self._df.columns:
            try:
                setattr(self, col_name, PublicOntologyField(self, col_name))
            # Some fields of an ontology (e.g. Gene) are not PublicOntology class attributes and must be skipped.
            except AttributeError:
                pass

    def __repr__(self) -> str:
        # fmt: off
        representation = (
            f"PublicOntology\n"
            f"Entity: {self.__class__.__name__}\n"
            f"Organism: {self.organism}\n"
            f"Source: {self.source}, {self.version}\n"
            f"#terms: {self._df.shape[0] if hasattr(self, '_df') else ''}\n\n"
            f"📖 .df(): ontology reference table\n"
            f"🔎 .lookup(): autocompletion of terms\n"
            f"🎯 .search(): free text search of terms\n"
            f"✅ .validate(): strictly validate values\n"
            f"🧐 .inspect(): full inspection of values\n"
            f"👽 .standardize(): convert to standardized names\n"
            f"🪜 .diff(): difference between two versions\n"
            f"🔗 .to_pronto(): Pronto.Ontology object"
        )
        # fmt: on
        return representation

    @property
    def organism(self):
        """The `name` of `Organism`."""
        return self._organism

    @property
    def source(self):
        """Name of the source."""
        return self._source

    @property
    def version(self):
        """Version of the source."""
        return self._version

    @property
    def fields(self) -> set:
        """All PublicOntology entity fields."""
        blacklist = {"include_id_prefixes"}
        fields = {
            field
            for field in vars(self)
            if not callable(getattr(self, field)) and not field.startswith("_")
        }
        return fields - blacklist

    def _download_ontology_file(self, localpath: Path, url: str, md5: str = "") -> None:
        """Download ontology source file to _local_ontology_path."""
        if not localpath.exists():
            logger.info(
                f"downloading {self.__class__.__name__} ontology source file..."
            )
            try:
                self._url_download(url, localpath)
            finally:
                # Only verify md5 if it's actually available from the sources.yaml file
                if len(md5) > 0:
                    if not verify_md5(localpath, md5):
                        logger.warning(
                            f"MD5 sum for {localpath} did not match {md5}. re-downloading..."
                        )
                        localpath.unlink()
                        self._url_download(url, localpath)

    def _fetch_sources(self) -> None:
        from ._display_sources import (
            display_available_sources,
            display_currently_used_sources,
        )

        def _subset_to_entity(df: pd.DataFrame, key: str):
            return df.loc[[key]] if isinstance(df.loc[key], pd.Series) else df.loc[key]

        self._current_sources = _subset_to_entity(
            display_currently_used_sources(), self.__class__.__name__
        )

        self._all_sources = _subset_to_entity(
            display_available_sources(), self.__class__.__name__
        )

    def _match_sources(
        self,
        ref_sources: pd.DataFrame,
        source: str | None = None,
        version: str | None = None,
        organism: str | None = None,
    ) -> dict[str, str]:
        """Match a source record base on passed organism, source and version."""
        lc = locals()

        # kwargs that are not None
        kwargs = {
            k: lc.get(k)
            for k in ["source", "version", "organism"]
            if lc.get(k) is not None
        }
        keys = list(kwargs.keys())

        # if 1 or 2 kwargs are specified, find the best match in currently used sources
        if (len(kwargs) == 1) or (len(kwargs) == 2):
            cond = ref_sources[keys[0]] == kwargs.get(keys[0])
            if len(kwargs) == 1:
                row = ref_sources[cond].head(1)
            else:
                # len(kwargs) == 2
                cond = cond.__and__(ref_sources[keys[1]] == kwargs.get(keys[1]))
                row = ref_sources[cond].head(1)
        else:
            # if no kwargs are passed, take the currently used source record
            if len(keys) == 0:
                curr = ref_sources.head(1).to_dict(orient="records")[0]
                kwargs = {
                    k: v
                    for k, v in curr.items()
                    if k in ["organism", "source", "version"]
                }
            # if all 3 kwargs are specified, match the record from currently used sources
            # do the same for the kwargs that obtained from default source to obtain url
            row = ref_sources[
                (ref_sources["organism"] == kwargs["organism"])
                & (ref_sources["source"] == kwargs["source"])
                & (ref_sources["version"] == kwargs["version"])
            ].head(1)

        # if no records matched the passed kwargs, raise error
        if row.shape[0] == 0:
            raise ValueError(
                f"No source is available with {kwargs}\nCheck"
                " `bionty.display_available_sources()`"
            )
        return row.to_dict(orient="records")[0]

    @check_dynamicdir_exists
    def _url_download(self, url: str, localpath: Path) -> None:
        """Download file from url to dynamicdir _local_ontology_path."""
        # Try to download from s3://bionty-assets
        s3_bionty_assets(
            filename=self._ontology_filename,  # type: ignore
            assets_base_url="s3://bionty-assets",
            localpath=localpath,
        )

        # If the file is not available, download from the url
        if not localpath.exists():
            logger.info(
                f"downloading {self.__class__.__name__} source file from: {url}"
            )
            _ = url_download(url, localpath)

    @check_datasetdir_exists
    def _set_file_paths(self) -> None:
        """Sets version, database and URL attributes for passed database and requested version."""
        self._url: str = self._source_record.get("url", "")
        self._md5: str = self._source_record.get("md5", "")

        # parquet file name, ontology source file name
        self._parquet_filename, self._ontology_filename = encode_filenames(
            organism=self.organism,
            source=self.source,
            version=self.version,
            entity=self,
        )
        self._local_parquet_path: Path = settings.dynamicdir / self._parquet_filename

        if self._url.endswith(".parquet"):  # user provide reference table as the url
            # no local ontology source file
            self._local_ontology_path = None  # type:ignore
            if not self._url.startswith("s3://bionty-assets/"):
                self._parquet_filename = None  # type:ignore
        else:
            self._local_ontology_path = settings.dynamicdir / self._ontology_filename

    def _get_default_field(self, field: PublicOntologyField | str | None = None) -> str:
        """Default to name field."""
        if field is None:
            if "name" in self._df.columns:
                field = "name"
            elif "symbol" in self._df.columns:
                field = "symbol"
            else:
                raise ValueError("Please specify a field!")
        field = str(field)
        if field not in self._df.columns:
            raise AssertionError(f"No {field} column exists!")
        return field

    def _load_df(self) -> pd.DataFrame:
        if self._parquet_filename is None:
            self._url_download(self._url, self._local_parquet_path)
        else:
            s3_bionty_assets(
                filename=self._parquet_filename,
                assets_base_url="s3://bionty-assets",
                localpath=self._local_parquet_path,
            )
        # If download is not possible, write a parquet file of the ontology df
        if not self._url.endswith("parquet"):
            if not self._local_parquet_path.exists():
                df = self.to_pronto().to_df(
                    source=self.source, include_id_prefixes=self.include_id_prefixes
                )
                df.to_parquet(self._local_parquet_path)

        # Loading the parquet file resets the index
        df = pd.read_parquet(self._local_parquet_path)
        return df

    def to_pronto(self):
        """The Pronto Ontology object.

        See: https://pronto.readthedocs.io/en/stable/api/pronto.Ontology.html
        """
        if self._local_ontology_path is None:
            logger.error(f"{self.__class__.__name__} has no Pronto Ontology object!")
            return
        else:
            self._download_ontology_file(
                localpath=self._local_ontology_path,
                url=self._url,
                md5=self._md5,
            )
            return Ontology(handle=self._local_ontology_path)

    def df(self) -> pd.DataFrame:
        """Pandas DataFrame of the ontology.

        Returns:
            A Pandas DataFrame of the ontology.

        Examples:
            >>> import bionty as bt
            >>> bt.Gene().df()
        """
        if "ontology_id" in self._df.columns:
            return self._df.set_index("ontology_id")
        else:
            return self._df

    def validate(
        self,
        values: Iterable,
        field: PublicOntologyField,
        *,
        mute: bool = False,
        **kwargs,
    ) -> np.ndarray:
        """Validate a list of values against a field of entity reference."""
        from lamin_utils._inspect import validate

        if isinstance(values, str):
            values = [values]

        field_values = self._df[str(field)]
        return validate(
            identifiers=values,
            field_values=field_values,
            case_sensitive=True,
            mute=mute,
            **kwargs,
        )

    def inspect(
        self,
        values: Iterable,
        field: PublicOntologyField,
        *,
        mute: bool = False,
        **kwargs,
    ) -> InspectResult:
        """Inspect a list of values against a field of entity reference.

        Args:
            values: Identifiers that will be checked against the field.
            field: The PublicOntologyField of the ontology to compare against.
                   Examples are 'ontology_id' to map against the source ID
                   or 'name' to map against the ontologies field names.
            return_df: Whether to return a Pandas DataFrame.
            mute: Whether to suppress logging. Defaults to False.
            kwargs: Used for backwards compatibility and return types.

        Returns:
            - A Dictionary of "validated" and "not_validated" identifiers
            - If `return_df`: A DataFrame indexed by identifiers with a boolean
                `__validated__` column indicating compliance validation.

        Examples:
            >>> import bionty as bt
            >>> public = bt.Gene()
            >>> gene_symbols = ["A1CF", "A1BG", "FANCD1", "FANCD20"]
            >>> public.inspect(gene_symbols, field=public.symbol)
        """
        from lamin_utils._inspect import inspect

        if isinstance(values, str):
            values = [values]

        return inspect(
            df=self._df,
            identifiers=values,
            field=str(field),
            mute=mute,
            **kwargs,
        )

    # unfortunately, the doc string here is duplicated with ORM.standardize
    def standardize(
        self,
        values: Iterable,
        field: PublicOntologyField | str | None = None,
        *,
        return_field: str = None,
        return_mapper: bool = False,
        case_sensitive: bool = False,
        mute: bool = False,
        keep: Literal["first", "last", False] = "first",
        synonyms_field: PublicOntologyField | str = "synonyms",
    ) -> dict[str, str] | list[str]:
        """Convert into standardized names.

        Args:
            values: `Iterable` Synonyms that will be standardized.
            field: `Optional[str]` The field representing the standardized names.
            return_field: `Optional[str]` The field to return. Defaults to field.
            return_mapper: `bool = False` If `True`, returns `{input_synonym1:
                standardized_name1}`.
            case_sensitive: `bool = False` Whether the mapping is case sensitive.
            keep: `Literal["first", "last", False] = "first"` When a synonym maps to
                multiple names, determines which duplicates to mark as
                `pd.DataFrame.duplicated`
            mute: Whether to mute logging. Defaults to False.

                    - "first": returns the first mapped standardized name
                    - "last": returns the last mapped standardized name
                    - `False`: returns all mapped standardized name
            synonyms_field: `str = "synonyms"` A field containing the concatenated synonyms.

        Returns:
            If `return_mapper` is `False`: a list of standardized names. Otherwise,
            a dictionary of mapped values with mappable synonyms as keys and
            standardized names as values.

        Examples:
            >>> import bionty as bt
            >>> public = bt.Gene()
            >>> gene_symbols = ["A1CF", "A1BG", "FANCD1", "FANCD20"]
            >>> standardized_symbols = public.standardize(gene_symbols, public.symbol)
        """
        from lamin_utils._standardize import standardize as map_synonyms

        if isinstance(values, str):
            values = [values]

        return map_synonyms(
            df=self._df,
            identifiers=values,
            field=self._get_default_field(field),
            return_field=self._get_default_field(return_field),
            return_mapper=return_mapper,
            case_sensitive=case_sensitive,
            mute=mute,
            keep=keep,
            synonyms_field=str(synonyms_field),
        )

    def map_synonyms(
        self,
        values: Iterable,
        *,
        return_mapper: bool = False,
        case_sensitive: bool = False,
        keep: Literal["first", "last", False] = "first",
        synonyms_field: PublicOntologyField | str = "synonyms",
        field: PublicOntologyField | str | None = None,
    ) -> dict[str, str] | list[str]:
        """Maps input synonyms to standardized names."""
        logger.warning("`map_synonyms()` is deprecated, use `.standardize()`!'")
        return self.standardize(
            values=values,
            return_mapper=return_mapper,
            case_sensitive=case_sensitive,
            keep=keep,
            synonyms_field=synonyms_field,
            field=field,
        )

    def lookup(self, field: PublicOntologyField | str | None = None) -> tuple:
        """An auto-complete object for a PublicOntology field.

        Args:
            field: The field to lookup the values for.
                   Defaults to 'name'.

        Returns:
            A NamedTuple of lookup information of the field values.

        Examples:
            >>> import bionty as bt
            >>> lookup = bt.CellType().lookup()
            >>> lookup.cd103_positive_dendritic_cell
            >>> lookup_dict = lookup.dict()
            >>> lookup['CD103-positive dendritic cell']
        """
        return Lookup(
            df=self._df,
            field=self._get_default_field(field),
            tuple_name=self.__class__.__name__,
            prefix="bt",
        ).lookup()

    def search(
        self,
        string: str,
        *,
        field: PublicOntologyField | str | None = None,
        limit: int | None = None,
        case_sensitive: bool = False,
        synonyms_field: PublicOntologyField | str | None = "synonyms",
    ):
        """Search a given string against a PublicOntology field.

        Args:
            string: The input string to match against the field values.
            field: The PublicOntologyField of the ontology the input string is matching against.
            top_hit: Return all entries ranked by matching ratios.
                If True, only return the top match.
                Defaults to False.
            limit: Maximum amount of top results to return.
                   If None, return all results.
                   Defaults to None.
            case_sensitive: Whether the match is case sensitive.
            synonyms_field: By default also search against the synonyms (If None, skips search).

        Returns:
            Ranked search results.

        Examples:
            >>> import bionty as bt
            >>> public = bt.CellType()
            >>> public.search("gamma delta T cell")
        """
        from lamin_utils._search import search

        return search(
            df=self._df,
            string=string,
            field=self._get_default_field(field),
            limit=limit,
            case_sensitive=case_sensitive,
            synonyms_field=str(synonyms_field),
        )

    def diff(
        self, compare_to: PublicOntology, **kwargs
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Determines a diff between two PublicOntology objects' ontologies.

        Args:
            compare_to: PublicOntology object that must be of the same class as the calling object.
            kwargs: Are passed to pd.DataFrame.compare()

        Returns:
            A tuple of two DataFrames:
            1. New entries.
            2. A pd.DataFrame.compare result which denotes all changes in `self` and `other`.

        Examples:
            >>> import bionty as bt
            >>> public_1 = bt.Disease(source="mondo", version="2023-04-04")
            >>> public_2 = bt.Disease(source="mondo", version="2023-04-04")
            >>> new_entries, modified_entries = public_1.diff(public_2)
            >>> print(new_entries.head())
            >>> print(modified_entries.head())
        """
        if type(self) is not type(compare_to):
            raise ValueError("Both PublicOntology objects must be of the same class.")

        if not self.source == compare_to.source:
            raise ValueError("Both PublicOntology objects must use the same source.")

        if self.version == compare_to.version:
            raise ValueError("The versions of the PublicOntology objects must differ.")

        # The 'parents' column (among potentially others) contain Numpy array values.
        # We transform them to tuples to determine the diff.
        def _convert_arrays_to_tuples(arr):  # pragma: no cover
            if isinstance(arr, np.ndarray):
                return tuple(arr)
            else:
                return arr

        for bt_obj in [self, compare_to]:
            for column in bt_obj.df().columns:
                if any(isinstance(val, np.ndarray) for val in bt_obj.df()[column]):
                    bt_obj._df[column] = bt_obj.df()[column].apply(
                        _convert_arrays_to_tuples
                    )

        # New entries
        new_entries = pd.concat([self.df(), compare_to.df()]).drop_duplicates(
            keep=False
        )

        # Changes in existing entries
        common_index = self.df().index.intersection(compare_to.df().index)
        self_df_common = self.df().loc[common_index]
        compare_to_df_common = compare_to.df().loc[common_index]
        modified_entries = self_df_common.compare(compare_to_df_common, **kwargs)

        logging.info(f"{len(new_entries)} new entries were added.")
        logging.info(f"{len(modified_entries)} entries were modified.")

        return new_entries, modified_entries


class PublicOntologyField:
    """Field of a PublicOntology model."""

    def __init__(self, parent: PublicOntology, name: str):
        self.parent = parent
        self.name = name

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name
